package com.scbfkj.uni.service;

import com.fasterxml.jackson.databind.JsonNode;
import com.jayway.jsonpath.JsonPath;
import com.scbfkj.uni.library.DataAliasGetUtil;
import com.scbfkj.uni.library.DataFormatUtil;
import com.scbfkj.uni.library.ScriptEngineUtil;
import com.scbfkj.uni.library.UniReturnUtil;
import com.scbfkj.uni.process.DataBase;
import com.scbfkj.uni.system.Config;
import jakarta.el.MethodNotFoundException;

import java.io.File;
import java.lang.reflect.Method;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.*;
import java.util.stream.Collectors;


public class DataProcessService {


    public static Map<String, Object> process(Map<String, Object> inData) throws Exception {

        Optional<String> serviceIdOpt = DataAliasGetUtil.getValue("serviceid", inData);
        if (Objects.isNull(inData) || inData.isEmpty() || serviceIdOpt.isEmpty())
            return UniReturnUtil.fail("服务编号不能为空");

        List<Map<String, Object>> resource = new ArrayList<>();
        resource.add(inData);
        String serviceId = serviceIdOpt.get();

        List<Map<String, Object>> algorithmLibraries = DataBase.query(Config.centerConnectionStr, "select * from algorithmlibrary where serviceid=?", Collections.singletonList(new Object[]{serviceId}));
        for (Map<String, Object> algorithmLibrary : algorithmLibraries) {
            resource.add(processByAlgorithm(algorithmLibrary, resource, null, null));
        }
        return UniReturnUtil.success(null);
    }

    /**
     * @param algorithmLibrary 算法配置
     * @param args 算法参数
     * @param filterColumns 数据库执行的列权限
     * @param filterLines 数据库执行的行权限
     * @return
     * @throws Exception
     */
    public static Map<String, Object> processByAlgorithm(Map<String, Object> algorithmLibrary, List<Map<String, Object>> args, List<String> filterColumns, Map<String, Object> filterLines) throws Exception {

        Object type = algorithmLibrary.get("algorithmtype");
        Object expression = algorithmLibrary.get("computingexpression");
        Object parameterSet = algorithmLibrary.get("parameterset");
        Object preConditions = algorithmLibrary.get("preconditions");
        Object executionNumber = algorithmLibrary.get("executionnumber");
        Object dataSourceId = algorithmLibrary.get("datasourceid");
        List<Map<String, Object>> datasourceList = DataBase.query(Config.centerConnectionStr, "select * from datasource where datasourceid=?", Collections.singletonList(new Object[]{dataSourceId}));
        Map<String, Object> datasource = datasourceList.get(0);
        if (Objects.nonNull(preConditions)) {
            String result = preConditionScript(preConditions.toString(), args);
            if (Objects.equals("1", result) || Objects.equals("2", result)) {
                return UniReturnUtil.success(result);
            }
        }
        List<Object> parameters = new ArrayList<>();
//        获取入参列表
        if (Objects.nonNull(parameterSet)) {
            HashMap<String, Object> source = new HashMap<>();
            source.put("args", args);
            source.put("algorithm", algorithmLibrary);
            source.put("datasource", datasource);
            parameters.add(getParams(parameterSet.toString(), source));
        }

        switch (type.toString()) {
//            java反射
            case "1" -> {
                Object connectset = datasource.get("connectset");
                JsonNode jsonNode = DataFormatUtil.toJsonNode(connectset);
                JsonNode path = jsonNode.get("path");
                if (Objects.nonNull(path)) {
                    File file = new File(System.getProperty("user.dir").concat(File.separator).concat("plugins").concat(File.separator).concat(path.asText()));
                    if (!file.exists()) {
                        throw new RuntimeException("外部文件加载不存在：".concat(System.getProperty("user.dir")).concat(File.separator).concat("plugins").concat(File.separator).concat(path.asText()));
                    }
                    Method addURLMethod = null;

                    boolean accessible = false;
                    try {
                        addURLMethod = URLClassLoader.class.getDeclaredMethod("addURL", URL.class);
                        URLClassLoader classLoader = (URLClassLoader) ClassLoader.getSystemClassLoader();
                        accessible = addURLMethod.canAccess(null); //获取方法的访问权限以便恢复:判读是否可以访问: true  代表可以访问
                        if (!accessible) { //判读是否可以访问，如果是false，那就必须setAccessible(true)
                            addURLMethod.setAccessible(true);//实际上是为了提效
                        }
                        addURLMethod.invoke(classLoader, file.toURI().toURL());//加载到系统中
                    } finally {
                        if (addURLMethod != null && !accessible) {
                            addURLMethod.setAccessible(false);//恢复原访问权限
                        }
                    }
                }
                JsonNode className = jsonNode.get("className");

                Class<?> classExample = Class.forName(className.asText()); //获取类实例
                Method javaMethod = null;
                Method closeMethod = null;
                Object classInstance = classExample.getConstructor().newInstance();//类实例接口 无参数构造
                for (Method currentMethod : classExample.getMethods()) {//循环所有方法
                    String methodName = currentMethod.getName();
                    if (methodName.equals(expression)) {
                        javaMethod = currentMethod;
                    }else if(methodName.equals("close")){
                        closeMethod=currentMethod;
                    }
                }
                if (Objects.isNull(javaMethod)) {
                    throw new MethodNotFoundException(expression + "： 没找到");
                }

                Object invoke = javaMethod.invoke(classInstance, parameters.toArray());
//                关闭 因为没有做实例的缓存 所以要关闭算法
                if(Objects.nonNull(closeMethod)){
                    closeMethod.invoke(classInstance);
                }
                return UniReturnUtil.success(invoke);
            }
//            JS表达式
            case "2" -> {
                return UniReturnUtil.success(ScriptEngineUtil.eval(expression.toString(), parameters.toArray()));
            }
//            数据库
            case "3" -> {
//                下放到Database中处理数据

            }
        }
        return UniReturnUtil.fail("");
    }

    /**
     * @param script  前置表达式 不能为空
     * @param args 表达式参数
     * @return
     * @throws Exception
     */
    private static String preConditionScript(String script, List<Map<String, Object>> args) throws Exception {
        return ScriptEngineUtil.eval(script, args);
    }

    /**
     * @param parameterSet 使用标准的jsonPath表达式 $.a.b[0].c,多个参数之间使用;;分隔
     * @param source jsonpath需要取值的数据源
     * @return
     */
    public static List<Object> getParams(String parameterSet, Object source) {
        String[] paths = parameterSet.split(";;");
        return Arrays.stream(paths).map(it -> JsonPath.read(source, it.trim())).toList();
    }

}
